/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.sm;

import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.Account;
import com.funambol.mailclient.sm.SyncClientConfig;
import com.funambol.mailclient.syncml.ContactSyncSource;
import com.funambol.mailclient.syncml.MailSyncSource;
import com.funambol.mailclient.syncml.MailSyncFilter;
import com.funambol.mailclient.syncml.MessageSyncSource;
import com.funambol.mailclient.MailFilter;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.controller.MailSyncListener;
import com.funambol.mailclient.ui.controller.ContactSyncListener;
import com.funambol.mailclient.ui.controller.MailSyncSingleMessageListener;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncConfig;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.spds.CompressedSyncException;
import com.funambol.syncml.spds.SyncManager;
import com.funambol.syncml.spds.SyncSource;
import com.funambol.util.Log;
import com.funambol.util.SyncListener;
import com.funambol.util.ThreadPool;

import com.funambol.mail.Message;

import java.io.IOException;


/**
 * A class that control and unify the sync process on the client side
 */
public class SyncClient implements Runnable {
    /**
     * SyncManager class declaration:
     * managed by a Singleton pattern in order to avoid memory leaks and
     * manage concurrent sync processes
     */
    private SyncManager sm;

    /** Source configuration */
    private SourceConfig sc;

    /** Instance of mail listener */
    private MailSyncListener mailListener;

    /** Instance of contactl listener */
    private ContactSyncListener contactListener;

    /** Instance of single message listener */
    private MailSyncSingleMessageListener singleMessageListener;

    /** Instance of this class*/
    private static SyncClient instance;

    // Constants that identify repositories that can be synchronized
    // This order is also the order in which the repositories are synchronized
    // when multiple repositories are requested for syncing. We always want the
    // messages to be first so the user has a faster response
    /** Constant that identifies MESSAGES repository */
    public static final int MESSAGES = 0;
    /** Constant that identifies CONTACTS repository */
    public static final int CONTACTS = 1;

    private static final int LAST_REP_ID = 1;

    /** Specifies which repositories need to be synchronized */
    private boolean repositoriesEnabled[];
    /** Specifies repositories synch mode */
    private int repositoriesSyncMode[];

    /** Specifies that a singleMsgSrc sync needs to be performed on this
     * source */
    private MessageSyncSource singleMsgSrc;

    /** Indicates if the SyncClient is busy (sync already in progress) */
    private boolean isBusy;

    /** Device User Agent used in HTTP headers */
    private String deviceUA;

    private int numRepositories;


    /**
     * Default (private) constructor:
     */
    private SyncClient() {

        reset();
    }

    ///////////////////////// Public methods /////////////////////////////////

    /**
     * This method returns the only SyncClient instance in the system. This
     * SyncClient is always associated to the default mail account that
     * is returned by the ConfigManager.
     */
    public static SyncClient getSyncClient() {

        if (instance == null) {
            instance = new SyncClient();
        }
        return instance;
    }



    /**
     * This method fires the synchronization of single repository.
     * This method cannot be invoked if a synchronization is already on going.
     * It is up to the client to verify this, for example by invoking the isBusy
     * method. If the method gets invoked when a synchronization is on going
     * then a SyncException is thrown.
     * This methods can only be invoke once a valid configuration has been
     * set, otherwise a ConfigException is thrown.
     * This method is non blocking and the synchronization is performed by a
     * separate thread. The client is notified trhough the SyncListener which is
     * linked to the SyncSource that is being synchronized.
     *
     * @param repository specifies the repository to sync
     *
     */
    public void sync(int repository) {

        resetRepState();
        repositoriesEnabled[repository]  = true;
        repositoriesSyncMode[repository] = -1;
        startThread();
    }

    /**
     * This method fires the synchronization of a single repository, allowing
     * the user to override the source default mode.
     * This method cannot be invoked if a synchronization is already on going.
     * It is up to the client to verify this, for example by invoking the isBusy
     * method. If the method gets invoked when a synchronization is on going
     * then a SyncException is thrown.
     * This methods can only be invoke once a valid configuration has been
     * set, otherwise a ConfigException is thrown.
     * This method is non blocking and the synchronization is performed by a
     * separate thread. The client is notified trhough the SyncListener which is
     * linked to the SyncSource that is being synchronized.
     *
     * @param repository specifies the repository to sync
     * @param mode synchronization mode
     *
     */
    public void sync(int repository, int mode) {

        resetRepState();
        repositoriesEnabled[repository]  = true;
        repositoriesSyncMode[repository] = mode;
        startThread();
    }

    /**
     * This method synchronizes all the known repositories, using sources
     * default sync mode.
     * This method cannot be invoked if a synchronization is already on going.
     * It is up to the client to verify this, for example by invoking the isBusy
     * method. If the method gets invoked when a synchronization is on going
     * then a SyncException is thrown.
     * This methods can only be invoke once a valid configuration has been
     * set, otherwise a ConfigException is thrown.
     * This method is non blocking and the synchronization is performed by a
     * separate thread. The client is notified trhough the SyncListener which is
     * linked to the SyncSource that is being synchronized.
     *
     */
    public void sync() {

        resetRepState();
        repositoriesEnabled[CONTACTS]  = true;
        repositoriesSyncMode[CONTACTS] = -1;
        repositoriesEnabled[MESSAGES]  = true;
        repositoriesSyncMode[MESSAGES] = -1;
        startThread();
    }

    /**
     * This method fires synchronizations for one or more repositories,
     * using sources default sync mode.
     * This method cannot be invoked if a synchronization is already on going.
     * It is up to the client to verify this, for example by invoking the isBusy
     * method. If the method gets invoked when a synchronization is on going
     * then a SyncException is thrown.
     * This methods can only be invoke once a valid configuration has been
     * set, otherwise a ConfigException is thrown.
     * This method is non blocking and the synchronization is performed by a
     * separate thread. The client is notified trhough the SyncListener which is
     * linked to the SyncSource that is being synchronized.
     *
     * @param repositories repositories to be synchronized
     */
    public void sync(int repositories[]) {

        resetRepState();
        // We do not catch ArrayOutOfBounds and if it occour we let it
        // raise
        for(int i=0; i<repositories.length;++i) {

            int repository = repositories[i];
            repositoriesEnabled[repository]  = true;
            repositoriesSyncMode[repository] = -1;
        }
        startThread();
    }


    /**
     * This method fires synchronizations for one or more repositories, allowing
     * the client to overried source default sync mode.
     * If multiple synchronizations are requested, then they are executed
     * sequentially and the order is unspecified.
     * These methods cannot be invoked if a synchronization is already on going.
     * It is up to the client to verify this, for example by invoking the isBusy
     * method. If the method gets invoked when a synchronization is on going
     * then a SyncException is thrown.
     * Theise methods can only be invoke once a valid configuration has been
     * set, otherwise a ConfigException is thrown.
     * This method is non blocking and the synchronization is performed by a
     * separate thread. The client is notified trhough the SyncListener which is
     * linked to the SyncSource that is being synchronized.
     *
     * @param repositories list of objects to be synchronized
     * @param mode syncronization mode
     *
     */
    public void sync(int repositories[], int modes[]) {

        resetRepState();
        // Synchronize all the objects requested by the client
        for(int i=0;i<repositories.length;++i) {
            int repository = repositories[i];
            int mode       = modes[i];

            repositoriesEnabled[repository]  = true;
            repositoriesSyncMode[repository] = mode;
        }
        startThread();
    }

    /**
     * Sync a single message. Messages are identified by an ID
     * which is composed by "folder" + "msgId". The client can specify if
     * attachments should be retrieved or not.
     * Today this method is mainly intended for the getFullMessage
     * functionality.
     * This method is non blocking and the synchronization is performed by a
     * separate thread. The client is notified trhough the SyncListener which is
     * linked to the SyncSource that is being synchronized.
     *
     * @param folder folder the message belongs to
     * @param luid   local id (LUID) of the message
     * @param withAttachments specifies if attachments must be included in the
     *                        message.
     *
     * @throws SyncException if there is already a synchronization in progess or
     *                       if something goes wrong during the sync
     * @throws ConfigException if there is no valid configuration for the
     *                         synchronization to be performed
     */
    public synchronized void syncSingleMessage(String folder, String luid,
            boolean withAttachments)
            throws ConfigException {

        // Perform the message sync
        MailSyncFilter inclusive = new MailSyncFilter(luid);
        inclusive.enableAttachmentsDownload(withAttachments);

        SourceConfig srcCfg = loadSourceConfig(SourceConfig.MAIL,
                SourceConfig.EMAIL_OBJECT_TYPE,
                "mail");
        singleMsgSrc = new MessageSyncSource(srcCfg, inclusive);
        singleMsgSrc.setListener(singleMessageListener);
        startThread();
    }

    /**
     * Set a listener on the mail sync source, used to signal the calling thread
     * of the relevant events occurred during mail synchronization.
     *
     * @param listener the SyncListener to set, or null to remove
     *                 the listener
     */
    public void setMailListener(MailSyncListener listener) {

        mailListener = listener;
    }

    /**
     * Set a listener on the contact sync source, used to signal the calling
     * thread of the relevant events occurred during contacts synchronization.
     *
     * @param listener the SyncListener to set, or null to remove
     *                 the listener
     */
    public void setContactListener(ContactSyncListener listener) {

        contactListener = listener;
    }

    /**
     * Set a listener on the single message sync source, used to signal the calling
     * thread of the relevant events occurred during single msg synchronization.
     *
     * @param listener the SyncListener to set, or null to remove
     *                 the listener
     */
    public void setSingleMessageListener(MailSyncSingleMessageListener listener) {

        singleMessageListener = listener;
    }

    /**
     * Factory method for the SourceConfig.
     * Loads and return the correct SouorceConfig depending on the source name.
     *
     * @param name the source name
     * @param type the source type
     * @param remoteUri the remore URI for the source
     */
    public SourceConfig loadSourceConfig(String name, String type,
            String remoteUri) {

        // create a new SourceConfig with the given paramenters
        sc = new SourceConfig(name, type, remoteUri);
        // The sources used by the mail client are all not encoded.
        sc.setEncoding(SyncSource.ENCODING_NONE);
        // Load the config from RMS.
        try {
            ConfigManager.load("source." + sc.getName(), sc);
            sc.setRemoteUri(remoteUri);
            Log.debug("Config for source" + sc.getName() + " loaded.");
        } catch(ConfigException ce) {
            Log.info("Config for [source: " + sc.getName() + ", uri: " +
                    remoteUri + "] not found, save default.");
            try {
                ConfigManager.save("source." + sc.getName(), sc);
            } catch (Exception e) {
                Log.error("loadSourceConfig, can't save default conf.");
                Log.error(" source: "+sc.getName()+" error: "+e.toString());
            }

        } catch (IOException ex) {
            ex.printStackTrace();
            Log.error("loadSourceConfig: "+ex.toString());
        }
        return sc;
    }

    /**
     * Save the given source config.
     *
     * @param sc the MailSourceConfig to save.
     *
     * @throws ConfigException if the config is not present
     * @throws IOException if an error occurred during the deserialization
     */
    public void saveSourceConfig()
    throws ConfigException, IOException {

        // The source config is stored in a record named "source.<sourcename>"
        ConfigManager.save("source." + sc.getName(), sc);
        Log.info("[SyncClient] config for source " + sc.getName() + " saved.");
    }

    /**
     * Check if a synchronization is in progress. SyncClient allows a sync at a
     * time. If one of the sync methods is invoked while synchronizing, then an
     * exception is thrown, but the client can check if a sync is in progress a
     * priori by using this method.
     * TODO FIXME: this mechanism is not thread safe
     */
    public boolean isBusy() {

        return isBusy;
    }


    /**
     * This method is the entry poin for the thread that performs the
     * synchronization. The user should never create the thread but he
     * should rely on the sync methods that spawn a thread for a non blocking
     * synchronization
     */
    public void run() {

        // If a synchronization is already in progress, then an exception is
        // thrown
        if (isBusy) {
            Log.info("[sync] sync already in progress.");
            throw new SyncException(
                    SyncException.CONCURRENCE_ERROR, "Sync in progress");
        }

        // Dump memory stats
        Log.stats("Beginning of SyncClient.run");

        numRepositories = 0;
        Log.info("- SyncClient: Start Sync -");
        long start = System.currentTimeMillis();
        // TODO FIXME: This is not thread safe
        isBusy = true;

        try {

            // At this point a configuration must be available, if not a
            // ConfigException gets propagated

            // Get the sync account
            Account a = ConfigManager.getConfig().getMailAccount();

            // We recreate a SyncManager at any sync, because we trade memory for
            // performance
            sm = getSyncManager(a);

            if (singleMsgSrc != null) {
                // In this case we must perform a single message sync, which is done
                // when the user wants to retrieve a full message
                Log.info("*** Separate thread for Sync of single message in progress ***");
                sm.sync(singleMsgSrc, SyncML.ALERT_CODE_ONE_WAY_FROM_SERVER);
                //save sync source configuration
                saveSourceConfig();
            } else {
                // This is the sync of complete repositories
                Log.info("*** Separate thread for Sync of repositories in progress ***");

                // Count the number of repositories to be synchronized
                for(int i=0;i<LAST_REP_ID+1;++i) {
                    if (repositoriesEnabled[i]) {
                        numRepositories++;
                    }
                }

                // Notify startMultiSession events
                if (numRepositories > 1) {
                    for(int i=0;i<LAST_REP_ID+1;++i) {
                        if (repositoriesEnabled[i]) {

                            if (i == MESSAGES) {
                                if (mailListener != null) {
                                    mailListener.startMultiSession(repositoriesEnabled);
                                }
                            } else if (i == CONTACTS) {
                                if (contactListener != null) {
                                    contactListener.startMultiSession(repositoriesEnabled);
                                }
                            }
                        }
                    }
                }

                // Synchronize the repositories
                for(int i=0;i<LAST_REP_ID+1;++i) {
                    if (repositoriesEnabled[i]) {
                        syncRepository(i, repositoriesSyncMode[i]);
                    }
                }

            }
        } catch (CompressedSyncException e) {
            Log.error(this, "Sync failed because compression failed");
            isBusy = false;
            // disable only-in-memory property for compression
            ConfigManager._tmpCompress = false;
            runUncompressedSync();
        } catch (SyncException se) {
            Log.error(this, "Sync failed because SyncManager failed");
            releaseResources();
            throw se;
        } catch(ConfigException ce) {
            Log.error(this, "Sync failed because source state cannot be saved");
            Log.error(this, ce.toString());
            releaseResources();
            throw new SyncException(
                    SyncException.STORAGE_ERROR,
                    "Error saving source state:" + ce.toString());
        } catch(IOException ioe) {
            Log.error(this, "Sync failed because source state cannot be saved");
            Log.error(this, ioe.toString());
            releaseResources();
            throw new SyncException(
                    SyncException.STORAGE_ERROR,
                    "Error saving source state:" + ioe.toString());
        } finally {
            // Notify endMultiSession events
            if (numRepositories > 1) {
                for(int i=0;i<LAST_REP_ID+1;++i) {
                    if (repositoriesEnabled[i]) {
                        if (i == MESSAGES) {
                            if (mailListener != null) {
                                mailListener.endMultiSession();
                            }
                        } else if (i == CONTACTS) {
                            if (contactListener != null) {
                                contactListener.endMultiSession();
                            }
                        }
                    }
                }
            }
            long syncTime = System.currentTimeMillis() - start;
            Log.info("Sync Time: " + (syncTime/1000));
            Log.info("- End Sync -");
            releaseResources();
            // Dump memory stats
            Log.stats("End of SyncClient.run");
        }
    }

    private void runUncompressedSync() {
        run();
    }

    private void releaseResources() {
        singleMsgSrc = null;
        // Reset the reference to the SyncManager to allow the GC to collect
        // memory if it needs to do so
        sm = null;
        isBusy = false;
        numRepositories = 0;
    }

    public String getDeviceUA() {
        if (deviceUA==null) {
            deviceUA = createUserAgent();
        }
        return deviceUA;
    }

    ///////////////////////// Private methods ////////////////////////////////

    /*
     * This method starts a synchronization thread. The syncs to be performed
     * have been set up by one of the corresponding sync methods.
     */
    private void startThread() {
        Log.debug("Thread for sync is starting");
        UIController.getThreadPool().startThread(this);
    }

    /**
     * This method resets the SyncClient and brings it to an empty state
     */
    private void reset() {
        sm           = null;
        sc           = null;
        mailListener = null;
        singleMsgSrc = null;
        isBusy       = false;

        repositoriesEnabled = new boolean[LAST_REP_ID + 1];
        repositoriesSyncMode = new int[LAST_REP_ID + 1];
        resetRepState();
    }

    /**
     * This method resets the repositories information. In particular the
     * information about which repositories need to be synced and their syncmode
     */
    private void resetRepState() {
        for(int i=0;i<LAST_REP_ID+1;++i) {
            repositoriesEnabled[i] = false;
            repositoriesSyncMode[i] = -1;
        }
    }

    /**
     * Set a filter on the underlying MailSyncSource. This method
     * must be called on an open connection only.
     * The filter is lost after a close() call.
     *
     * @param MailFilter the filter to apply, or null to remove it
     * @throws MailClientException if the connection is not open
     */
    private void setFilter(MailSyncSource ss, MailFilter filter)
    throws SyncException {

        // Nullify the filter if the source is present, ignore otherwise
        if (filter == null) {
            if(ss != null) {
                ss.setFilter(null);
            }
            return;
        }

        // Throw an exception if the source is not initialized yet
        if(ss == null) {
            Log.error(this, "Can't set a filter on a closed connection.");
            throw new SyncException(SyncException.CLIENT_ERROR,"Must be connected to set a filter");
        }

        // Set the source filter
        MailSyncFilter sf = new MailSyncFilter(filter);
        ss.setFilter(sf);
    }

    /**
     * Helper method used to synchronize a single repository, allowing the
     * client to override the sync mode defined in the sync source.
     * An appropriate SyncSource is created, depending on the repository that
     * needs to be synchronized. Then the synchronization is fired by invoking
     * the SyncManager sync method. If everything is OK then the configuration
     * is saved.
     *
     * @param repository identifies the repository to be synchronized
     * @param mode allows overwriting of the source default mode (-1 implies the
     *             usage of the source default mode)
     *
     * @throws SyncException if something goes wrong during synchronization
     * @throws ConfigException if the SyncManager cannot be instantiated because
     *                         the configuration is invalid.
     */
    private void syncRepository(int repository, int mode)
    throws SyncException, ConfigException, IOException {

        SourceConfig srcCfg;
        SyncSource   source;

        // Get the sync account
        Account a = ConfigManager.getConfig().getMailAccount();
        if (repository == MESSAGES) {
            Log.info("SyncClient: synchronizing messages ");
            
            srcCfg = loadSourceConfig(SourceConfig.MAIL,
                    SourceConfig.EMAIL_OBJECT_TYPE,
                    a.getRemoteURI());
            MailSyncSource ms = new MailSyncSource(srcCfg);
            if (mailListener != null)
                ms.setListener(mailListener);
            source = ms;
            setFilter(ms, ConfigManager.getConfig().getMailFilter());
        } else if (repository == CONTACTS) {
            Log.info("SyncClient: synchronizing contacts ");
            srcCfg = loadSourceConfig(SourceConfig.CONTACT,
                    SourceConfig.VCARD_TYPE,
                    a.getPimRemoteURI());
            ContactSyncSource ss = new ContactSyncSource(srcCfg);
            if (contactListener != null)
                ss.setListener(contactListener);
            source = ss;
        } else {
            Log.error(this,"Unknown object to sync" );
            throw new SyncException(SyncException.CLIENT_ERROR, "Unknown object to sync");
        }

        Log.info("Source: "+source.getName());
        Log.info("SyncMode: "+source.getSyncMode());

        //do sync
        if (mode != -1) {
            sm.sync(source, mode);
        } else {
            sm.sync(source);
        }
        
        //save sync source configuration
        saveSourceConfig();
    }


    private SyncManager getSyncManager(Account a) {
        // Init a SyncConfig with the parameters get from Account
        SyncConfig syncConfig = new SyncConfig();

        syncConfig.syncUrl = a.getUrl();
        syncConfig.userName = a.getUser();
        syncConfig.password = a.getPassword();
        syncConfig.userAgent = createUserAgent();

        //#ifdef http_resp.force_cookies
//#         syncConfig.forceCookies = true;
        //#endif

        Log.info("HTTP Device-Agent header: " + syncConfig.userAgent);

        // Get SyncClientConfig for Syncml specific parameters
        SyncClientConfig conf = new SyncClientConfig();

        try {
            ConfigManager.load(SyncClientConfig.CONFIGNAME, conf);

        } catch (Exception ce) {
            Log.info("[getSyncManager] Error reading SyncClientConfig, using defaults.");
            // Get a new device Id, it's the first use of the client
            // on the device.
            conf.generateDeviceId();
        }

        // Set device id in the devinfo config class
        syncConfig.deviceConfig.devID = conf.getDeviceId();

        syncConfig.lastServerUrl = conf.getLastSyncUrl();
        conf.setLastSyncUrl(syncConfig.syncUrl);

        // TODO: check exception handling!!!!!!!!!
        try {
            // set if the communications with the server must be compressed
            //syncConfig.compress = ConfigManager.getConfig().isCompress();
            syncConfig.compress = ConfigManager._tmpCompress;
            // Get a new instance of SyncManager
            sm = new SyncManager(syncConfig);

            ConfigManager.save(SyncClientConfig.CONFIGNAME, conf);

        } catch (Exception e) {
            Log.error("[getSyncManager] Error saving SyncClientConfig Config: " + e.toString());
        }
        return sm;
    }

    // Creates UserAgent to use in HTTP requests
    private String createUserAgent() {

        StringBuffer sbAgent = new StringBuffer();
        
        sbAgent.append(ConfigManager.getClientName() + " Java Mail");
        sbAgent.append(" v. ").append(ConfigManager.getClientVersion());
        
        return sbAgent.toString();
    }

}

